#!/usr/bin/env python3

import argparse
import glob
import os
import sys
from concurrent.futures import ProcessPoolExecutor

import pandas as pd
import pyarrow.parquet as pq
import pyranges as pr


def read_bed_file(bed_file):
    """Read BED file and keep only the first three columns."""
    bed_df = pd.read_csv(bed_file, sep="\t", header=None, comment="#")
    bed_df = bed_df.iloc[:, :3]
    bed_df.columns = ["Chromosome", "Start", "End"]
    return pr.PyRanges(bed_df)


def filter_motif_data(motif_df, quality=None, pvalue=None, qvalue=None):
    """Filter the motif DataFrame by quality, p-value, and q-value thresholds."""
    if quality and "quality" in motif_df.columns:
        motif_df = motif_df[motif_df["quality"].isin(quality)]

    if pvalue and "p-value" in motif_df.columns:
        motif_df = motif_df[motif_df["p-value"] <= pvalue]

    if qvalue and "q-value" in motif_df.columns:
        motif_df = motif_df[motif_df["q-value"] <= qvalue]

    if motif_df.empty:
        return None

    return motif_df[["sequence_name", "start", "stop"]].drop_duplicates()


def process_motif_dir(motif_dir, regions_pr, quality=None, pvalue=None, qvalue=None):
    """Process a single motif directory to compute overlaps."""
    motif_name = os.path.basename(motif_dir).split("=")[1]
    motif_parquet_files = glob.glob(os.path.join(motif_dir, "*.parquet"))

    # Read and concatenate Parquet files
    motif_dfs = []
    for file in motif_parquet_files:
        try:
            motif_df = pq.read_table(file).to_pandas()
            motif_dfs.append(motif_df)
        except Exception as e:
            print(f"Error reading {file}: {e}")
            continue

    if not motif_dfs:
        print(f"No data for motif {motif_name}. Skipping.")
        return None

    motif_df = pd.concat(motif_dfs, ignore_index=True)
    motif_df = filter_motif_data(motif_df, quality=quality, pvalue=pvalue, qvalue=qvalue)

    if motif_df is None:
        print(f"No data after filtering for {motif_name}. Skipping.")
        return None

    # Prepare motif intervals
    motif_df["Chromosome"] = motif_df["sequence_name"]
    motif_df["Start"] = motif_df["start"] - 1  # Convert to 0-based
    motif_df["End"] = motif_df["stop"]
    motif_pr = pr.PyRanges(motif_df[["Chromosome", "Start", "End"]])

    # Compute overlaps
    overlaps = motif_pr.join(regions_pr)
    count = len(overlaps)

    return {"motif_name": motif_name, "count": count}


def compute_motif_overlaps(bed_file, fimo_parquet_dir, quality=None, pvalue=None, qvalue=None, threads=4):
    """Main function to compute overlaps using multiple processes."""
    regions_pr = read_bed_file(bed_file)
    motif_dirs = glob.glob(os.path.join(fimo_parquet_dir, "motif_name=*"))

    if not motif_dirs:
        print(f"No motif directories found in {fimo_parquet_dir}")
        return None

    # Process motif directories in parallel
    results = []
    with ProcessPoolExecutor(max_workers=threads) as executor:
        futures = [
            executor.submit(process_motif_dir, motif_dir, regions_pr, quality, pvalue, qvalue)
            for motif_dir in motif_dirs
        ]
        for future in futures:
            result = future.result()
            if result:
                results.append(result)

    # Collect results
    results_df = pd.DataFrame(results).sort_values("motif_name")
    return results_df


def main():
    parser = argparse.ArgumentParser(description="Compute motif intersections with regions of interest.")
    parser.add_argument("bed_file", help="Input BED file with regions of interest.")
    parser.add_argument("fimo_parquet_dir", help="Directory containing motif data in Parquet format.")
    parser.add_argument(
        "--output",
        help="Output file name. If not provided, results will be written to stdout.",
    )
    parser.add_argument(
        "--output-format",
        choices=["tsv", "parquet"],
        default="tsv",
        help="Output file format.",
    )
    parser.add_argument(
        "--quality",
        choices=["A", "B", "C"],
        nargs="+",
        help="Filter motifs by quality. Accepts one or more of A, B, C.",
    )
    parser.add_argument("--pvalue", type=float, help="Filter motifs with p-value <= specified value.")
    parser.add_argument("--qvalue", type=float, help="Filter motifs with q-value <= specified value.")
    parser.add_argument("--threads", type=int, default=4, help="Number of parallel threads to use.")

    args = parser.parse_args()

    # Call the compute function
    results_df = compute_motif_overlaps(
        bed_file=args.bed_file,
        fimo_parquet_dir=args.fimo_parquet_dir,
        quality=args.quality,
        pvalue=args.pvalue,
        qvalue=args.qvalue,
        threads=args.threads,
    )

    if results_df is None or results_df.empty:
        print("No intersection results to display.")
        sys.exit(0)

    # Output results
    if args.output:
        if args.output_format == "tsv":
            results_df.to_csv(args.output, sep="\t", index=False)
        elif args.output_format == "parquet":
            results_df.to_parquet(args.output, index=False)
        print(f"Results saved to {args.output}")
    else:
        if args.output_format == "tsv":
            print(results_df.to_csv(sep="\t", index=False))
        elif args.output_format == "parquet":
            print("Cannot write Parquet format to stdout. Please specify an output file using '--output'.")
            sys.exit(1)


if __name__ == "__main__":
    main()
